import os
import sys
import subprocess
import pickle
from .utils import format_return, insert_list, docker_error, get_mdi_standard, get_compose_path, get_package_path, get_mdimechanic_yaml, writelines_as_bytes, write_as_bytes

# Paths to enter each identified node
node_paths = { "@DEFAULT": "" }

# Paths associated with the edges for the node graph
node_edge_paths = [ ("@DEFAULT", "") ]

# Use MPI for the tests
use_mpi = False

def docker_up( base_path, docker_path ):
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )

    # Create the docker environment
    docker_env = os.environ
    docker_env['MDIMECH_WORKDIR'] = base_path
    docker_env['MDIMECH_PACKAGEDIR'] = get_package_path()
    docker_env['MDIMECH_ENGINE_NAME'] = mdimechanic_yaml['docker']['image_name'] + ":dev"

    # Run "docker-compose up -d"
    up_proc = subprocess.Popen( ["docker-compose", "up", "-d"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                cwd=docker_path, env=docker_env )
    up_tup = up_proc.communicate()
    up_out = format_return(up_tup[0])
    up_err = format_return(up_tup[1])
    return up_proc.returncode

def docker_down( base_path, docker_path ):
    # Create the docker environment
    docker_env = os.environ
    docker_env['MDIMECH_WORKDIR'] = base_path
    docker_env['MDIMECH_PACKAGEDIR'] = get_package_path()

    # Run "docker-compose down"
    down_proc = subprocess.Popen( ["docker-compose", "down"],
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                  cwd=docker_path, env=docker_env )
    down_tup = down_proc.communicate()
    down_out = format_return(down_tup[0])
    down_err = format_return(down_tup[1])
    return down_proc.returncode


def test_command( base_path, command, nrecv, recv_type, nsend, send_type ):
    global use_mpi
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )

    # Remove any leftover files from previous runs of min_driver.py
    #base_path = get_base_path()
    dat_file = os.path.join( base_path, ".mdimechanic", ".temp", "min_driver.dat" )
    err_file = os.path.join( base_path, ".mdimechanic", ".temp", "min_driver.err" )
    if os.path.exists( dat_file ):
        os.remove( dat_file )
    if os.path.exists( err_file ):
        os.remove( err_file )

    if use_mpi:
        mdi_driver_options = "-role DRIVER -name driver -method MPI"
        mdi_engine_options = "-role ENGINE -name TESTCODE -method MPI"
        docker_path = get_compose_path( "mpi" )
    else:
        mdi_driver_options = "-role DRIVER -name driver -method TCP -port 8021"
        mdi_engine_options = "-role ENGINE -name TESTCODE -method TCP -hostname mdi_mechanic -port 8021"
        docker_path = None
        if 'gpu' in mdimechanic_yaml['docker']:
            docker_path = get_compose_path( "tcp_nvidia" )
        else:
            docker_path = get_compose_path( "tcp" )

    # Create the script for MDI Mechanic
    docker_file = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_mechanic.sh" )
    docker_lines = [ "#!/bin/bash\n",
                     "\n",
                     "# Exit if any command fails\n",
                     "\n",
                     "cd /MDI_Mechanic/mdimechanic/drivers\n",
                     "python min_driver.py \\\n"
    ]
    if command is not None:
        docker_lines.append( "   -command \'" + str(command) + "\' \\\n" )
    if nrecv is not None:
        docker_lines.append( "   -nreceive \'" + str(nrecv) + "\' \\\n" )
    if recv_type is not None:
        docker_lines.append( "   -rtype \'" + str(recv_type) + "\' \\\n" )
    if nsend is not None:
        docker_lines.append( "   -nsend \'" + str(nsend) + "\' \\\n" )
    if send_type is not None:
        docker_lines.append( "   -stype \'" + str(send_type) + "\' \\\n" )
    docker_lines.append( "   -mdi \"" + str(mdi_driver_options) + "\"\n" )
    os.makedirs(os.path.dirname(docker_file), exist_ok=True)
    writelines_as_bytes( docker_lines, docker_file )

    script_lines = mdimechanic_yaml['engine_tests']['script']
    script = "#!/bin/bash\nset -e\ncd /repo\n"
    script += "export MDI_OPTIONS=\'" + str(mdi_engine_options) + "\'\n"
    for line in script_lines:
        script += line + '\n'

    # Write the script to run the test
    script_path = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_engine.sh" )
    os.makedirs(os.path.dirname(script_path), exist_ok=True)
    write_as_bytes( script, script_path )



    if use_mpi:

        # Start the docker container
        if docker_up( base_path, docker_path ) != 0:
            raise Exception("Unable to start docker-compose")

        # Create the docker environment
        docker_env = os.environ
        docker_env['MDIMECH_WORKDIR'] = base_path
        docker_env['MDIMECH_PACKAGEDIR'] = get_package_path()
        docker_env['MDIMECH_ENGINE_NAME'] = mdimechanic_yaml['docker']['image_name'] + ":dev"

        # Run "docker-compose exec"
        exec_proc = subprocess.Popen( ["docker-compose", "exec", "-T", "--user", "mpiuser", "mdi_mechanic", "mpiexec", "-app", "/MDI_Mechanic/mdimechanic/docker/mpi/mdi_appfile"],
                                      stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                      cwd=docker_path, env=docker_env )
        exec_tup = exec_proc.communicate()
        exec_out = format_return(exec_tup[0])
        exec_err = format_return(exec_tup[1])
        if exec_proc.returncode != 0:
            print("FAILED", flush=True)
            docker_down( base_path, docker_path )
            return False

    else:

        # Create the docker environment
        docker_env = os.environ
        docker_env['MDIMECH_WORKDIR'] = base_path
        docker_env['MDIMECH_PACKAGEDIR'] = get_package_path()
        docker_env['MDIMECH_ENGINE_NAME'] = mdimechanic_yaml['docker']['image_name'] + ":dev"

        # Run the docker container
        up_proc = subprocess.Popen( ["docker-compose", "up", "--exit-code-from", "mdi_mechanic", "--abort-on-container-exit"],
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                    cwd=docker_path, env=docker_env )
        up_tup = up_proc.communicate()
        up_out = format_return(up_tup[0])
        up_err = format_return(up_tup[1])
        if up_proc.returncode != 0:
            print("FAILED", flush=True)
            docker_down( base_path, docker_path )
            return False



    if docker_down( base_path, docker_path ) != 0:
        print("FAILED", flush=True)
        return False


    print("WORKED", flush=True)
    return True



def find_nodes( base_path ):
    global node_paths
    global node_edge_paths

    mdimechanic_yaml = get_mdimechanic_yaml( base_path )
    
    # List of all node commands in the MDI Standard
    command_list = []
    commands = None
    
    #base_path = get_base_path()
    standard = get_mdi_standard( base_path )
    commands = standard['commands']
    
    for command in commands:
        if command[0] == '@' and command != '@':
            command_list.append( command )
    ordered_commands = sorted( command_list )
    ####################
    #return
    ####################

    # Check which of the MDI Standard commands work from the @DEFAULT node
    for command in ordered_commands:
        command_works = test_command( base_path, command, None, None, None, None )
        if command_works:
            node_paths[command] = command
            node_edge_paths.append( (command, command) )
    
    # From the nodes that have currently been identified, attempt to use the "@" command to identify more nodes
    print("Searching for supported nodes", flush=True)
    original_nodes = []
    for node in node_paths.keys():
        original_nodes.append(node)
    ordered_nodes = sorted( original_nodes )

    # Determine the maximum distance to search for new nodes
    node_search_distance = 5
    if 'engine_tests' in mdimechanic_yaml:
        if 'node_search_distance' in mdimechanic_yaml['engine_tests']:
            node_search_distance = mdimechanic_yaml['engine_tests']['node_search_distance']

    for node in ordered_nodes:
        for ii in range(node_search_distance):
            new_path = node_paths[node]
            for jj in range(ii+1):
                new_path += " @"
            command = new_path + " <@"
            print("Checking for node at: " + str(new_path), end=" ")
            command_works = test_command( base_path, command, "MDI_COMMAND_LENGTH", "MDI_CHAR", None, None )
        
            # Read the name of the node
            node_name = None
            dat_file = os.path.join( base_path, ".mdimechanic", ".temp", "min_driver.dat" )
            err_file = os.path.join( base_path, ".mdimechanic", ".temp", "min_driver.err" )
            if os.path.isfile( dat_file ):
                with open( dat_file, "r") as f:
                    node_name = f.read()
            err_value = None
            if os.path.isfile( err_file ):
                with open( err_file, "r") as f:
                    err_value = f.read()
                
            if node_name is not None and not node_name in node_paths.keys():
                node_paths[node_name] = new_path

            # Check whether this should be added to the node graph
            if node_name is not None:
                split_path = new_path.split()
                include = True
                for node_edge in node_edge_paths:
                    if node_edge[0] == node_name:
                        if node_edge[1] == '':
                            path = ['',]
                        else:
                            path = node_edge[1].split()
                        if path[0] == split_path[0]:
                            include = False

                if include:
                    node_edge_paths.append( (node_name, new_path) )

    print("Completed search for nodes.", flush=True)
    print("Found the following nodes: " + str(node_paths.keys()) )

def write_supported_commands( base_path ):
    global node_paths

    mdimechanic_yaml = get_mdimechanic_yaml( base_path )
    
    # List of all commands in the MDI Standard
    command_list = []
    commands = None
    
    #base_path = get_base_path()
    standard = get_mdi_standard( base_path )
    commands = standard['commands']
    
    for command in commands.keys():
        values = commands[command]
        command_list.append( command )
    ordered_commands = sorted( command_list )

    # Identify all supported nodes, and find a path to them
    find_nodes( base_path )
    ordered_nodes = sorted( node_paths.keys() )
    
    # Write the README.md section that lists all supported commands
    command_sec = []

    # Write the section header
    command_sec.append( "## Supported Commands\n" )
    command_sec.append( "\n" )
    header_line = "| "
    for node in ordered_nodes:
        header_line += "| " + str(node) + " "
    header_line += "|\n"
    command_sec.append( header_line )
    header_line = "| ------------- "
    for node in ordered_nodes:
        header_line += "| ------------- "
    header_line += "|\n"
    command_sec.append( header_line )

    # Write the list of supported commands
    for command in ordered_commands:
        nrecv = None
        recv_type = None
        nsend = None
        send_type = None
        print("---------------------------------------")
        print("Testing command: " + str(command))
        if commands[command] is not None and 'recv' in commands[command].keys():
            nrecv = commands[command]['recv']['count']
            recv_type = commands[command]['recv']['datatype']
        if commands[command] is not None and 'send' in commands[command].keys():
            nsend = commands[command]['send']['count']
            send_type = commands[command]['send']['datatype']
        
        line = "| " + str(command) + " "
        for node in ordered_nodes:
            command_with_path = node_paths[node] + " " + command
            padded_string = str(node).ljust(20, '.')
            print(padded_string, end=" ")

            # If there is a list of commands to test, only test the listed commands:
            run_test = True
            if 'engine_tests' in mdimechanic_yaml:
                if 'tested_commands' in mdimechanic_yaml['engine_tests']:
                    if not command in mdimechanic_yaml['engine_tests']['tested_commands']:
                        run_test = False

            if run_test:
                command_works = test_command( base_path, command_with_path, nrecv, recv_type, nsend, send_type )
            else:
                command_works = False
                print("SKIPPED", flush=True)
        
            if command_works:
                # Display a bright green box
                command_status = "![command](report/badges/box-brightgreen.svg)"
            else:
                # Display a light gray box
                command_status = "![command](report/badges/box-lightgray.svg)"

            line += "| " + str(command_status) + " "
        line += "|\n"
        command_sec.append( line )

    # Replace all ">" or "<" symbols with Markdown escape sequences
    for iline in range(len(command_sec)):
        line = command_sec[iline]
        line = line.replace(">", "&gt;")
        line = line.replace("<", "&lt;")
        command_sec[iline] = line
        
    return command_sec

def node_graph( base_path ):
    global node_edge_paths

    print("*********************************************")
    print("* Creating node graph                       *")
    print("*********************************************")

    print("node_edge_paths: " + str(node_edge_paths))

    package_path = get_package_path()

    nodes = {}
    edges = []
    for edge_path in node_edge_paths:
        name = edge_path[0]
        path = edge_path[1].split()
        if '@' in path:
            parent_cluster = str(path[0]) + '_'
            if not parent_cluster in nodes.keys():
                nodes[ parent_cluster ] = str(name)
                edges.append( ( path[0], parent_cluster ) )
            else:
                nodes[ parent_cluster ] += '\n' + str(name)
        else:
            nodes[ name ] = name
            if name != '@DEFAULT':
                edges.append( ( '@DEFAULT', name ) )
    print("nodes: " + str(nodes))

    # Save the graph data to a file
    graph_data = { 'nodes': nodes,
                   'edges': edges }
    graph_file = os.path.join( base_path, ".mdimechanic", ".temp", "graph.pickle")
    os.makedirs(os.path.dirname(graph_file), exist_ok=True)
    with open(graph_file, 'wb') as handle:
        pickle.dump(graph_data, handle, protocol=min(pickle.HIGHEST_PROTOCOL, 4))

    # Graph script
    graph_script = '''
import os
import pickle
from graphviz import Digraph

# Read the data required to make the graph
#graph_file = os.path.join( \'/repo\', \'.mdimechanic\', \'.temp\', \'graph.pickle\')
graph_file = \'/repo/.mdimechanic/.temp/graph.pickle\'

with open(graph_file, 'rb') as handle:
    data = pickle.load(handle)
nodes = data['nodes']
edges = data['edges']

dot = Digraph(comment='Node Report', format='svg')

node_list = []
for node in nodes.keys():
    node_list.append( node )
ordered_nodes = sorted( node_list )
    
for node in ordered_nodes:
    dot.node( node, nodes[ node ], shape='box', margin='0.1' )

for edge in edges:
    dot.edge( edge[0], edge[1] )

#graph_path = os.path.join( base_path, "report", "graphs", "node-report.gv" )
graph_path = '/repo/report/graphs/node-report.gv'
dot.render( graph_path )
'''

    # Render the graph within a docker image, so that it is consistent across machines
    graph_proc = subprocess.Popen( ["docker", "run", "--rm",
                                    "-v", str(base_path) + ":/repo",
                                    "-v", str(package_path) + ":/MDI_Mechanic",
                                    "mdi_mechanic/mdi_mechanic",
                                    "bash", "-c",
                                    "cd /MDI_Mechanic/mdimechanic/utils && python -c \"" + graph_script + "\""],
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    graph_tup = graph_proc.communicate()
    if graph_proc.returncode != 0:
        docker_error( graph_tup, "Graph process returned an error." )

def analyze_nodes( base_path ):
    #package_path = get_package_path()

    # Read the README.md file
    readme_path = os.path.join(base_path, "README.md")
    with open(readme_path, "r") as file:
        readme = file.readlines()

    # Check the README.md file for any comments for travis
    for iline in range(len(readme)):
        line = readme[iline]
        sline = line.split()
        if len(sline) > 0 and sline[0] == '[travis]:':
            instruction = sline[3]

            if instruction == "supported_commands":
                # Need to insert a list of supported commands here
                command_sec = write_supported_commands( base_path )
                insert_list( readme, command_sec, iline )

    # Write the updates to the README file
    temp_file = os.path.join( base_path, '.mdimechanic', '.temp', 'README.temp')
    os.makedirs(os.path.dirname(temp_file), exist_ok=True)
    writelines_as_bytes( readme, temp_file )

    # Create the node graph
    node_graph( base_path )



import os
import pickle
from graphviz import Digraph

def make_graph( base_path ):
    # Read the data required to make the graph
    graph_file = os.path.join( base_path, "MDI_Mechanic", ".temp", "graph.pickle")
    with open(graph_file, 'rb') as handle:
        data = pickle.load(handle)
    nodes = data['nodes']
    edges = data['edges']

    dot = Digraph(comment='Node Report', format='svg')

    node_list = []
    for node in nodes.keys():
        node_list.append( node )
    ordered_nodes = sorted( node_list )
    
    for node in ordered_nodes:
        dot.node( node, nodes[ node ], shape='box', margin='0.1' )

    for edge in edges:
        dot.edge( edge[0], edge[1] )

    graph_path = os.path.join( base_path, "report", "graphs", "node-report.gv" )
    dot.render( graph_path )

if __name__ == "__main__":
    make_graph()



import os
import subprocess
import shutil
from .utils.utils import format_return, insert_list, docker_error, get_mdi_standard, get_compose_path, get_package_path, get_mdimechanic_yaml, writelines_as_bytes, write_as_bytes


def test_driver( driver_name, base_path ):
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )

    # Get the path to the docker-compose file
    docker_path = None
    if 'gpu' in mdimechanic_yaml['docker']:
        docker_path = get_compose_path( "nvidia_tcp" )
    else:
        docker_path = get_compose_path( "tcp" )

    # Write the run script for MDI Mechanic
    docker_file = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_mechanic.sh" )
    docker_lines = [ "#!/bin/bash -l\n",
                     "set -e\n",
                     "\n",
                     "cd /repo\n"]

    driver_script = mdimechanic_yaml['test_drivers'][driver_name]['script']
    for line in driver_script:
        docker_lines.append( line + '\n' )
    os.makedirs(os.path.dirname(docker_file), exist_ok=True)
    writelines_as_bytes( docker_lines, docker_file )

    # Write the run script for the engine
    # NOTE: NEED TO LOOP OVER ALL AVAIALBLE TEST SCRIPTS
    script_lines = mdimechanic_yaml['engine_tests']['script']
    script = "#!/bin/bash\nset -e\ncd /repo\n"
    script += "export MDI_OPTIONS=\'-role ENGINE -name TESTCODE -method TCP -hostname mdi_mechanic -port 8021\'\n"
    for line in script_lines:
        script += line + '\n'

    # Write the script to run the test
    script_path = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_engine.sh" )
    os.makedirs(os.path.dirname(script_path), exist_ok=True)
    write_as_bytes( script, script_path )

    # Create the docker environment
    docker_env = os.environ
    docker_env['MDIMECH_WORKDIR'] = base_path
    docker_env['MDIMECH_PACKAGEDIR'] = get_package_path()
    docker_env['MDIMECH_ENGINE_NAME'] = mdimechanic_yaml['docker']['image_name'] + ":dev"

    # Run "docker-compose up"
    up_proc = subprocess.Popen( ["docker-compose", "up", "--exit-code-from", "mdi_mechanic", "--abort-on-container-exit"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                cwd=docker_path, env=docker_env )
    up_tup = up_proc.communicate()
    up_out = format_return(up_tup[0])
    up_err = format_return(up_tup[1])


    # Run "docker-compose down"
    down_proc = subprocess.Popen( ["docker-compose", "down"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                  cwd=docker_path, env=docker_env )
    down_tup = down_proc.communicate()
    down_out = format_return(down_tup[0])
    down_err = format_return(down_tup[1])

    if up_proc.returncode != 0:
        docker_error( up_tup, "Driver test returned non-zero exit code." )

    elif down_proc.returncode != 0:
        docker_error( down_tup, "Driver test returned non-zero exit code on docker down." )

    else:
        print("====================================================")
        print("================ Output from Docker ================")
        print("====================================================")
        print(up_out)
        print("====================================================")
        print("============== End Output from Docker ==============")
        print("====================================================")



import os
import pickle
import sys
import yaml

def parse_mdi_standard( ):
    #package_path = get_package_path()
    #file_path = os.path.dirname(os.path.realpath(__file__))
    #package_path = os.path.dirname( os.path.dirname( file_path ) )
    #base_path = os.path.dirname( os.path.dirname( os.path.dirname( file_path ) ) )
    #standard_yaml_path = os.path.join( base_path,"MDI_Mechanic","mdi_standard.yaml")
    #standard_yaml_path = os.path.join( package_path, "mdimechanic", "mdi_standard.yaml")
    #standard_file = os.path.join( base_path, "MDI_Mechanic", ".temp", "standard.pickle" )

    # This is only executed within Docker, so the paths should NOT use os.path.join
    standard_yaml_path = "/MDI_Mechanic/mdimechanic/mdi_standard.yaml"
    standard_file = "/repo/.mdimechanic/.temp/standard.pickle"

    # Read the yaml file
    with open(standard_yaml_path, "r") as yaml_file:
        standard = yaml.load(yaml_file, Loader=yaml.FullLoader)

    # Dump the data using pickle
    #print("FILE: " + str(base_path))
    os.makedirs(os.path.dirname(standard_file), exist_ok=True)
    with open(standard_file, 'wb') as handle:
        pickle.dump(standard, handle, protocol=min(pickle.HIGHEST_PROTOCOL, 4))

if __name__ == "__main__":
    parse_mdi_standard( )



import os
import pickle
import subprocess
import yaml

def get_base_path():
    # Get the base directory
    raise Exception("Should not call get_base_path anymore")
    file_path = os.path.dirname(os.path.realpath(__file__))
    base_path = os.path.dirname( os.path.dirname( os.path.dirname( file_path ) ) )
    return base_path

def get_package_path():
    file_path = os.path.dirname(os.path.realpath(__file__))
    package_path = os.path.dirname( os.path.dirname( file_path ) )
    #package_path = os.path.dirname( file_path )
    return package_path

def get_compose_path( method ):
    package_path = get_package_path()
    compose_path = os.path.join( package_path, "mdimechanic", "docker", method )
    return compose_path

def format_return(input_string):
    my_string = input_string.decode('utf-8')

    # remove any \r special characters, which sometimes are added on Windows
    my_string = my_string.replace('\r','')

    return my_string

def insert_list( original_list, insert_list, pos ):
    for ielement in range(len(insert_list)):
        element = insert_list[ielement]
        original_list.insert( pos + ielement + 1, element )

def docker_error( docker_tup, error_message ):
    docker_out = format_return(docker_tup[0])
    docker_err = format_return(docker_tup[1])
    print("-------- BEGIN DOCKER OUTPUT --------")
    print( str(docker_out) )
    print("-------- END DOCKER OUTPUT ----------")
    print("-------- BEGIN DOCKER ERROR ---------")
    print( str(docker_err) )
    print("-------- END DOCKER ERROR -----------")
    raise Exception(error_message)

def get_mdi_standard( base_path ):
    # Path to the file where the standard will be written
    package_path = get_package_path()
    standard_file = os.path.join( base_path, ".mdimechanic", ".temp", "standard.pickle" )
    
    parse_proc = subprocess.Popen( ["docker", "run", "--rm",
                                    "-v", str(base_path) + ":/repo",
                                    "-v", str(package_path) + ":/MDI_Mechanic",
                                    "mdi_mechanic/mdi_mechanic",
                                    "bash", "-c",
                                    "cd /MDI_Mechanic/mdimechanic/utils && python parse_standard.py"],
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    parse_tup = parse_proc.communicate()
    if parse_proc.returncode != 0:
        docker_error( parse_tup, "Parse process returned an error." )

    with open(standard_file, 'rb') as handle:
        standard = pickle.load(handle)
    return standard

def get_mdimechanic_yaml( base_path ):
    yaml_path = os.path.join( base_path, "mdimechanic.yml" )
    with open(yaml_path, "r") as yaml_file:
        mdimechanic_yaml = yaml.load(yaml_file, Loader=yaml.FullLoader)
    return mdimechanic_yaml

def write_as_bytes( text, file ):
    # Write as bytes, to ensure line returns work correctly on Windows
    with open(file, "wb") as opened_file:
        opened_file.write( bytes(text, "UTF-8") )

def writelines_as_bytes( lines, file ):
    # Write as bytes, to ensure line returns work correctly on Windows
    text = ""
    for line in lines:
        text += line
    with open(file, "wb") as opened_file:
        opened_file.write( bytes(text, "UTF-8") )



import mdi
from mdi import MDI_NAME_LENGTH, MDI_COMMAND_LENGTH
import pandas as pd
import re
import sys

command = None
nreceive = None
rtype = None
nsend = None
stype = None

iarg = 1
while iarg < len(sys.argv):
    arg = sys.argv[iarg]

    if arg == "-mdi":
        # Initialize MDI
        if len(sys.argv) <= iarg+1:
            raise Exception("Argument to -mdi option not found")
        mdi.MDI_Init(sys.argv[iarg+1])
        iarg += 1
    elif arg == "-command":
        # Set the command
        if len(sys.argv) <= iarg+1:
            raise Exception("Argument to -command option not found")
        command_input = sys.argv[iarg+1]
        iarg += 1
    elif arg == "-nreceive":
        # Set the number of elements to receive
        if len(sys.argv) <= iarg+1:
            raise Exception("Argument to -nreceive option not found")
        nreceive = sys.argv[iarg+1]
        iarg += 1
    elif arg == "-nsend":
        # Set the number of elements to send
        if len(sys.argv) <= iarg+1:
            raise Exception("Argument to -nsend option not found")
        nsend = sys.argv[iarg+1]
        iarg += 1
    elif arg == "-rtype":
        # Set the type of elements to receive
        if len(sys.argv) <= iarg+1:
            raise Exception("Argument to -rtype option not found")
        rtype = sys.argv[iarg+1]
        iarg += 1
    elif arg == "-stype":
        # Set the type of elements to receive
        if len(sys.argv) <= iarg+1:
            raise Exception("Argument to -stype option not found")
        stype = sys.argv[iarg+1]
        iarg += 1
    else:
        raise Exception("Unrecognized argument")

    iarg += 1

# Connect to the engine
comm = mdi.MDI_Accept_Communicator()

# Get the name of the engine, which will be checked and verified at the end
mdi.MDI_Send_Command("<NAME", comm)
initial_name = mdi.MDI_Recv(mdi.MDI_NAME_LENGTH, mdi.MDI_CHAR, comm)

recv_type = None
if nreceive is not None:
    # Get the number of elements to receive
    nreceive_split = re.split("\+|\-|\*\*|\*|\/\/|\/|\%| ",nreceive)
    for word in nreceive_split:
        if len(word) > 0 and word[0] == '<':
            # This is a command, so send it to the engine
            # Assume that the command receives a single integer
            mdi.MDI_Send_Command(word, comm)
            value = mdi.MDI_Recv(1, mdi.MDI_INT, comm)
            
            nreceive = nreceive.replace(word, str(value), 1)
    
    recv_num = pd.eval(nreceive)
    
    # Confirm that the receive type is valid
    if rtype == "MDI_CHAR":
        recv_type = mdi.MDI_CHAR
    elif rtype == "MDI_INT":
        recv_type = mdi.MDI_INT
    elif rtype == "MDI_DOUBLE":
        recv_type = mdi.MDI_DOUBLE
    elif rtype == "MDI_BYTE":
        recv_type = mdi.MDI_BYTE
    else:
        raise Exception("Invalid receive type")
        
send_type = None
if nsend is not None:
    # Get the number of elements to send
    nsend_split = re.split("\+|\-|\*\*|\*|\/\/|\/|\%| ",nsend)
    for word in nsend_split:
        if len(word) > 0 and word[0] == '<':
            # This is a command, so send it to the engine
            # Assume that the command receives a single integer
            mdi.MDI_Send_Command(word, comm)
            value = mdi.MDI_Recv(1, mdi.MDI_INT, comm)
            
            nsend = nsend.replace(word, str(value), 1)
    
    send_num = pd.eval(nsend)
    
    # Confirm that the receive type is valid
    if stype == "MDI_CHAR":
        send_type = mdi.MDI_CHAR
    elif stype == "MDI_INT":
        send_type = mdi.MDI_INT
    elif stype == "MDI_DOUBLE":
        send_type = mdi.MDI_DOUBLE
    elif stype == "MDI_BYTE":
        send_type = mdi.MDI_BYTE
    else:
        raise Exception("Invalid send type")


# Send all commands
command_string = str(command_input)
command_list = command_string.split()

# Send the command(s) to be tested
for command in command_list:
    mdi.MDI_Send_Command(command, comm)

# Send or receive any data associated with the final command
if nreceive is not None:
    recv_data = mdi.MDI_Recv(recv_num, recv_type, comm)
    
    # Write the received data to a file 
    f = open("/repo/.mdimechanic/.temp/min_driver.dat", "w")
    f.write(str(recv_data))
    f.close()
if nsend is not None:
    if send_type == mdi.MDI_INT:
        data = [ 0 for i in range(send_num) ]
    elif send_type == mdi.MDI_DOUBLE:
        data = [ 0.0 for i in range(send_num) ]
    elif send_type == mdi.MDI_CHAR:
        data = ""
        for i in range(send_num):
            data += " "
    else:
        raise Exception("Invalid send type")
    mdi.MDI_Send(data, send_num, send_type, comm)

# Verify that the engine is still responsive
mdi.MDI_Send_Command("<NAME", comm)
final_name = mdi.MDI_Recv(mdi.MDI_NAME_LENGTH, mdi.MDI_CHAR, comm)
assert initial_name == final_name

# Some nodes might not support the "EXIT" command, so write a file indicating success now
# Write the received data to a file
f = open("/repo/.mdimechanic/.temp/min_driver.err", "w")
f.write("0")
f.close()

mdi.MDI_Send_Command("EXIT", comm)



import os
import subprocess
import shutil
from .utils import utils as ut


def start( image_name, base_path ):
    mdimechanic_yaml = ut.get_mdimechanic_yaml( base_path )

    # Name of the image to run
    if image_name is None:
        # The user did not supply an image name, so try using a ":dev" tag
        image_name = mdimechanic_yaml['docker']['image_name'] + ":dev"

    # Find the location of .gitconfig file
    gitconfig_line = ""
    host_home_dir = os.path.expanduser('~')
    linux_location = os.path.join(host_home_dir, '.gitconfig')
    found_file = os.path.exists( linux_location )
    if found_file:
        gitconfig_line = " -v " + linux_location + ":/root/.gitconfig"
    else: # Check if this is Windows
        windows_location = os.path.join( str( os.environ['USERPROFILE'] ), ".gitconfig" )
        found_file = os.path.exists( windows_location )
        if found_file:
            gitconfig_line = " -v " + windows_location + ":/root/.gitconfig"
    if not found_file:
        print("WARNING: MDI Mechanic was unable to locate a .gitconfig file.  Git will not be fully functional within this interactive session.")

    # Find the location of the .ssh directory
    ssh_line = ""
    host_home_dir = os.path.expanduser('~')
    linux_location = os.path.join(host_home_dir, '.ssh')
    found_file = os.path.exists( linux_location )
    if found_file:
        ssh_line = " -v " + linux_location + ":/root/.ssh"
    else: # Check if this is Windows
        windows_location = os.path.join( str( os.environ['USERPROFILE'] ), ".ssh" )
        found_file = os.path.exists( windows_location )
        if found_file:
            ssh_line = " -v " + windows_location + ":/root/.ssh"
    if not found_file:
        print("WARNING: MDI Mechanic was unable to locate a .ssh directory.  Git will not be fully functional within this interactive session.")

    # This script will be executed on entry to the image
    interactive_entry_script = '''#!/bin/bash -l
set -e
if [ ! -d "/repo" ]; then
  ln -s /mdi_shared /repo
fi
cd /repo
bash
'''

    # Write the entry script into the mounted volume
    interactive_entry_path = os.path.join( base_path, "docker", ".temp", "interactive_entry.sh" )
    os.makedirs(os.path.dirname(interactive_entry_path), exist_ok=True)
    ut.write_as_bytes( interactive_entry_script, interactive_entry_path )

    # Construct the command line to launch docker interactively
    run_line = "docker run --rm"
    run_line += " -v " + str( base_path ) + ":/mdi_shared"
    run_line += gitconfig_line
    run_line += ssh_line
    if 'gpu' in mdimechanic_yaml['docker']:
        run_line += " --gpus all"
    if 'extra_launch_options' in mdimechanic_yaml['docker']:
        run_line += " " + str(mdimechanic_yaml['docker']['extra_launch_options'])
    run_line += " -it " + str(image_name) + " bash /mdi_shared/docker/.temp/interactive_entry.sh"
    print("Interactive session run command: " + str(run_line))
    os.system(run_line)



import os
import subprocess
import shutil
from .utils import format_return, insert_list, docker_error, get_mdi_standard, get_compose_path, get_package_path, get_mdimechanic_yaml, writelines_as_bytes, write_as_bytes



def test_validate( base_path ):
    # Get the base directory
    package_path = get_package_path()

    # Get the MDI YAML
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )

    # Read the yaml script for validating the engine build
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )
    validate_engine_lines = mdimechanic_yaml['docker']['validate_engine']
    validate_engine_script = "#!/bin/bash\nset -e\ncd /repo\n"
    for line in validate_engine_lines:
        validate_engine_script += line + '\n'

    # Write the script to validate the engine build
    validate_script_path = os.path.join( base_path, ".mdimechanic", ".temp", "validate_engine.sh" )
    os.makedirs(os.path.dirname(validate_script_path), exist_ok=True)
    write_as_bytes( validate_engine_script, validate_script_path )

    # Run the test
    test_proc = subprocess.Popen( ["docker", "run", "--rm",
                                   "-v", str(base_path) + ":/repo",
                                   "-v", str(package_path) + ":/MDI_Mechanic",
                                   mdimechanic_yaml['docker']['image_name'] + ":dev",
                                   "bash", "-l", "-c",
                                   "cd /repo && bash .mdimechanic/.temp/validate_engine.sh"],
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    test_tup = test_proc.communicate()
    test_out = format_return(test_tup[0])
    test_err = format_return(test_tup[1])

    if test_proc.returncode != 0:
        docker_error( test_tup, "Build validation script returned non-zero value." )


def test_min( base_path ):
    # Get the path to the docker-compose file
    docker_path = get_compose_path( "tcp" )

    # Write the run script for MDI Mechanic
    docker_file = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_mechanic.sh" )
    docker_lines = [ "#!/bin/bash -l\n",
                     "\n",
                     "# Exit if any command fails\n",
                     "\n",
                     "cd /MDI_Mechanic/mdimechanic/drivers\n",
                     "python min_driver.py -command \'<NAME\' -nreceive \'MDI_NAME_LENGTH\' -rtype \'MDI_CHAR\' -mdi \'-role DRIVER -name driver -method TCP -port 8021\'\n"]
    os.makedirs(os.path.dirname(docker_file), exist_ok=True)
    writelines_as_bytes( docker_lines, docker_file )

    # Write the run script for the engine
    # NOTE: NEED TO LOOP OVER ALL AVAIALBLE TEST SCRIPTS
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )
    script_lines = mdimechanic_yaml['engine_tests']['script']
    script = "#!/bin/bash -l\nset -e\ncd /repo\n"
    script += "export MDI_OPTIONS=\'-role ENGINE -name TESTCODE -method TCP -hostname mdi_mechanic -port 8021\'\n"
    for line in script_lines:
        script += line + '\n'

    # Write the script to run the test
    script_path = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_engine.sh" )
    os.makedirs(os.path.dirname(script_path), exist_ok=True)
    write_as_bytes( script, script_path )

    # Prepare the working directory
    #src_path = os.path.join( base_path, "user", "mdi_tests", "test1" )
    #dst_path = os.path.join( base_path, "user", "mdi_tests", ".work" )
    #if os.path.isdir( dst_path ):
    #    shutil.rmtree( dst_path )
    #shutil.copytree( src_path, dst_path )

    # Create the docker environment
    docker_env = os.environ
    docker_env['MDIMECH_WORKDIR'] = base_path
    docker_env['MDIMECH_PACKAGEDIR'] = get_package_path()
    docker_env['MDIMECH_ENGINE_NAME'] = mdimechanic_yaml['docker']['image_name'] + ":dev"

    # Run "docker-compose up"
    up_proc = subprocess.Popen( ["docker-compose", "up", "--exit-code-from", "mdi_mechanic", "--abort-on-container-exit"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                cwd=docker_path, env=docker_env )
    up_tup = up_proc.communicate()
    up_out = format_return(up_tup[0])
    up_err = format_return(up_tup[1])

    # Run "docker-compose down"
    down_proc = subprocess.Popen( ["docker-compose", "down"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                  cwd=docker_path, env=docker_env )
    down_tup = down_proc.communicate()
    down_out = format_return(down_tup[0])
    down_err = format_return(down_tup[1])

    if up_proc.returncode != 0:
        docker_error( up_tup, "Minimal MDI functionality test returned non-zero exit code." )

    if down_proc.returncode != 0:
        docker_error( down_tup, "Minimal MDI functionality test returned non-zero exit code on docker down." )

def test_unsupported( base_path ):
    # Get the path to the docker-compose file
    docker_path = get_compose_path( "tcp" )

    # Write the run script for MDI Mechanic
    docker_file = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_mechanic.sh" )
    docker_lines = [ "#!/bin/bash\n",
                     "\n",
                     "# Exit if any command fails\n",
                     "\n",
                     "cd /MDI_Mechanic/mdimechanic/drivers\n",
                     "python min_driver.py -command \'<UNSUPPORTED\' -nreceive \'MDI_NAME_LENGTH\' -rtype \'MDI_CHAR\' -mdi \'-role DRIVER -name driver -method TCP -port 8021\'\n"]
    os.makedirs(os.path.dirname(docker_file), exist_ok=True)
    writelines_as_bytes( docker_lines, docker_file )

    # Write the run script for the engine
    #docker_file = str(base_path) + '/MDI_Mechanic/.temp/docker_mdi_engine.sh'
    #docker_lines = [ "#!/bin/bash\n",
    #                 "\n",
    #                 "# Exit if any command fails\n",
    #                 "\n",
    #                 "cd /repo/user/mdi_tests/.work\n",
    #                 "export MDI_OPTIONS=\'-role ENGINE -name TESTCODE -method TCP -hostname mdi_mechanic -port 8021\'\n",
    #                 "./run.sh\n"]
    #os.makedirs(os.path.dirname(docker_file), exist_ok=True)
    # NOTE: NEED TO LOOP OVER ALL AVAIALBLE TEST SCRIPTS
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )
    script_lines = mdimechanic_yaml['engine_tests']['script']
    script = "#!/bin/bash -l\nset -e\ncd /repo\n"
    script += "export MDI_OPTIONS=\'-role ENGINE -name TESTCODE -method TCP -hostname mdi_mechanic -port 8021\'\n"
    for line in script_lines:
        script += line + '\n'

    # Write the script to run the test
    script_path = os.path.join( base_path, ".mdimechanic", ".temp", "docker_mdi_engine.sh" )
    os.makedirs(os.path.dirname(script_path), exist_ok=True)
    write_as_bytes( script, script_path )
        
    # Prepare the working directory
    #src_path = os.path.join( base_path, "user", "mdi_tests", "test1" )
    #dst_path = os.path.join( base_path, "user", "mdi_tests", ".work" )
    #if os.path.isdir( dst_path ):
    #    shutil.rmtree( dst_path )
    #shutil.copytree( src_path, dst_path )

    # Create the docker environment
    docker_env = os.environ
    docker_env['MDIMECH_WORKDIR'] = base_path
    docker_env['MDIMECH_PACKAGEDIR'] = get_package_path()
    docker_env['MDIMECH_ENGINE_NAME'] = mdimechanic_yaml['docker']['image_name'] + ":dev"

    # Run "docker-compose up"
    up_proc = subprocess.Popen( ["docker-compose", "up", "--exit-code-from", "mdi_mechanic", "--abort-on-container-exit"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                cwd=docker_path, env=docker_env )
    up_tup = up_proc.communicate()
    up_out = format_return(up_tup[0])
    up_err = format_return(up_tup[1])

    # Run "docker-compose down"
    down_proc = subprocess.Popen( ["docker-compose", "down"],
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                  cwd=docker_path, env=docker_env )
    down_tup = down_proc.communicate()
    down_out = format_return(down_tup[0])
    down_err = format_return(down_tup[1])

    #assert up_proc.returncode != 0
    #assert down_proc.returncode == 0

    if up_proc.returncode == 0:
        docker_error( up_tup, "Test for correct error functionality returned zero exit code." )

    if down_proc.returncode != 0:
        docker_error( down_tup, "Test for correct error functionality returned non-zero exit code on docker down." )



"""
mdi_mechanic.py
A tool for testing and developing MDI-enabled projects.

Handles the primary functions
"""

import os
import shutil
from . import cmd_report
from . import cmd_build
from . import cmd_run
from . import cmd_rundriver
from . import cmd_interactive
from .utils import utils as ut

def command_report():
    print("Starting a report")
    report_dir = os.getcwd()
    cmd_report.generate_report( report_dir )

def command_build( args ):
    print("Starting the installation")

    build_type = args.pop("build_type")

    if ( build_type != "dev" and build_type != "release"):
        raise Exception("Error: invalid argument to --type option.")

    report_dir = os.getcwd()
    cmd_build.install_all( report_dir, build_type )

def command_startproject( args ):
    print("Starting a new MDI project")
    project_type = args.pop("projecttype")

    base_path = os.getcwd()
    package_path = ut.get_package_path()

    if project_type == "enginereport":
        project_path = os.path.join( package_path, "mdimechanic", "projects", "enginereport" )

        print("Starting an engine report project")

        yml_path = os.path.join( base_path, "mdimechanic.yml" )
        docker_path = os.path.join( base_path, "docker" )
        gitignore_path = os.path.join( base_path, ".gitignore" )
        if os.path.exists( yml_path ) or os.path.exists( docker_path ):
            raise Exception("This already appears to be an MDI project")

        yml_source = os.path.join( project_path, "mdimechanic.yml" )
        docker_source = os.path.join( project_path, "docker" )
        workflow_source = os.path.join( project_path, ".github" )
        gitignore_source = os.path.join( project_path, ".gitignore" )
        shutil.copyfile( yml_source, yml_path )
        shutil.copytree( docker_source, docker_path )
        shutil.copyfile( gitignore_source, gitignore_path )

    else:
        raise Exception("Unrecognized project type.")

def command_run( args ):
    run_dir = os.getcwd()
    print("Running a custom calculation with MDI Mechanic.")
    script_name = args.pop("script_name")

    if script_name is None:
        raise Exception("Error: --name argument was not provided.")

    # Test the driver
    try:
        cmd_run.run( script_name, run_dir )
        print("Success: The driver ran to completion.")
    except:
        raise Exception("Error: The script did not complete successfully.")

def command_rundriver( args ):
    report_dir = os.getcwd()
    print("Running an MDI test driver.")
    driver_name = args.pop("driver_name")

    # Test the driver
    try:
        cmd_rundriver.test_driver( driver_name, report_dir )
        print("Success: The driver ran to completion.")
    except:
        raise Exception("Error: The test driver did not complete successfully.")

def command_interactive( args ):
    run_dir = os.getcwd()
    image_name = args.pop("image_name")

    try:
        cmd_interactive.start( image_name, run_dir )
        print("Success: Interactive session completed successfully.")
    except:
        raise Exception("Error: Interactive session did not complete successfully.")



import os
import subprocess
from shutil import copyfile
from .utils import tests as mtests
from .utils import node_analysis as na
from .utils import reset_report as rr
from .utils import utils as ut

# Generate the report
def generate_report( base_path ):

    # Path to this file
    #file_path = os.path.dirname(os.path.realpath(__file__))

    # Path to the top-level directory
    #base_path = os.path.dirname( os.path.dirname( file_path ) )

    # Reset the report
    rr.reset_report( base_path )

    # Ensure that there are no orphaned containers / networks running
    try:
        compose_path = ut.get_compose_path( "tcp" )
        down_proc = subprocess.Popen( ["docker-compose", "down"],
                                      stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                      cwd=compose_path)
        down_tup = down_proc.communicate()
    except:
        raise Exception("Error: Unable to remove orphaned containers.")
    try:
        compose_path = ut.get_compose_path( "tcp_nvidia" )
        down_proc = subprocess.Popen( ["docker-compose", "down"],
                                      stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                      cwd=compose_path)
        down_tup = down_proc.communicate()
    except:
        raise Exception("Error: Unable to remove orphaned containers.")

    # Verify that the engine has been built / installed correctly
    try:
        mtests.test_validate( base_path )
        print("Success: Able to verify that the engine was built")
        src_location = os.path.join(base_path, "report", "badges", "-working-success.svg")
        dst_location = os.path.join(base_path, "report", "dynamic_badges", "step_engine_build.svg")
        copyfile(src_location, dst_location)
    except:
        raise Exception("Error: Unable to verify that the engine was built.")

    # Verify that the engine supports minimalistic MDI functionality
    try:
        mtests.test_min( base_path )
        print("Success: Engine passed minimal MDI functionality test.")
        src_location = os.path.join(base_path, "report", "badges", "-working-success.svg")
        dst_location = os.path.join(base_path, "report", "dynamic_badges", "step_min_engine.svg")
        copyfile(src_location, dst_location)

    except:
        raise Exception("Error: Engine failed minimal MDI functionality test.")

    # Check if the engine correctly errors upon receiving an unsupported command
    try:
        mtests.test_unsupported( base_path )
        print("Success: Engine errors out upon receiving an unsupported command.")
        src_location = os.path.join(base_path, "report", "badges", "-working-success.svg")
        dst_location = os.path.join(base_path, "report", "dynamic_badges", "step_unsupported.svg")
        copyfile(src_location, dst_location)
    except:
        raise Exception("Error: Engine does not error out upon receiving an unsupported command.")

    # Perform the node analysis
    try:
        na.analyze_nodes( base_path )
        print("Success: Detected MDI nodes.")

        # Copy the success badge for this step
        src_location = os.path.join(base_path, "report", "badges", "-working-success.svg")
        dst_location = os.path.join(base_path, "report", "dynamic_badges", "step_mdi_nodes.svg")
        copyfile(src_location, dst_location)

    except:
        raise Exception("Error: Unable to detect MDI nodes.")

    # Prepend the Travis badge to README.md
    readme_path = os.path.join(base_path, ".mdimechanic", ".temp", "README.temp")
    badge_path = os.path.join(base_path, ".mdimechanic", "ci_badge.md")
    badge = None
    readme = None
    try:
        with open(badge_path, 'r') as original: badge = original.read()
    except FileNotFoundError:
        badge = ""
    with open(readme_path, 'r') as original: readme = original.read()
    with open(readme_path, 'w') as modified: modified.write(badge + readme)
    
    # Copy README.temp over README.md
    src_location = os.path.join(base_path, ".mdimechanic", ".temp", "README.temp")
    dst_location = os.path.join(base_path, "README.md")
    copyfile(src_location, dst_location)



import os
import shutil
from .utils import get_package_path, get_mdimechanic_yaml, insert_list, writelines_as_bytes

def reset_report( base_path ):
    # Path to this file
    #file_path = os.path.dirname(os.path.realpath(__file__))

    # Path to the top-level directory
    #base_path = os.path.dirname( os.path.dirname( os.path.dirname( file_path ) ) )

    # Get the MDI YAML
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )

    package_path = get_package_path()

    # Remove the old report
    report_path = os.path.join(base_path, "report")
    if os.path.isdir( report_path ):
        shutil.rmtree( report_path )

    # Copy the base report
    src_location = os.path.join( package_path, "mdimechanic", "report", "base_report" )
    dst_location = os.path.join( base_path, "report")
    shutil.copytree( src_location, dst_location )

    # Get the README.md file locations
    src_location = os.path.join( package_path, "mdimechanic", "report", "README.base" )
    dst_location = os.path.join( base_path, "README.md")

    # Get header information for the README
    code_name = mdimechanic_yaml['code_name']
    readme_data = ["# MDI Mechanic " + str(code_name) + " report\n",
                    "\n",
                    "This repo presents test results for the MDI interface implementation in the " + str(code_name) + " code.\n",
                    "\n"]

    # Get the body information for the README
    with open( src_location, 'r' ) as base_file:
        body = base_file.readlines()

    # Add any prepended information from the mdimechanic.yml file
    prepend_iline = 0
    prepend_data = []
    for iline in range(len(body)):
        line = body[iline]
        sline = line.split()
        if len(sline) > 3:
            if sline[0] == '[yaml]:':
                instruction = sline[3]
                if instruction == "prepend":
                    prepend_iline = iline
    if 'report' in mdimechanic_yaml:
        if 'prepend' in mdimechanic_yaml['report']:
            prepend_data = mdimechanic_yaml['report']['prepend']
            for iline in range(len(prepend_data)):
                if prepend_data[iline] == None:
                    prepend_data[iline] = ""
                prepend_data[iline] = prepend_data[iline] + "\n"
            insert_list( body, prepend_data, prepend_iline)

    # Combine the header data with the body of the README
    readme_data.extend( body )

    # Write the README.md file
    #with open( dst_location, 'w' ) as new_file:
    #    new_file.writelines( readme_data )
    writelines_as_bytes( readme_data, dst_location )

if __name__ == "__main__":
    reset_report()



"""
Provides a CLI for MDI_Mechanic
"""

import argparse
import sys
from . import commands

def parse_args():
    parser = argparse.ArgumentParser(description="A CLI for the MDI_Mechanic.")

    subparsers = parser.add_subparsers(dest="command")

    build = subparsers.add_parser("build", help="Build containers for MDI_Mechanic and the engine.")
    #build.add_argument("location", type=str, help="The location where the engine is located.")
    build.add_argument("--type", dest='build_type',
                     type=str,
                     default="dev",
                     help="Type of build. (\"dev\" [Default] or \"release\")")

    startproject = subparsers.add_parser("startproject", help="Start a new MDI project.")
    startproject.add_argument("--enginereport", dest='projecttype', action='store_const',
                              const="enginereport", default=None,
                              help="Start an engine report project.")

    report = subparsers.add_parser("report", help="Create a report for the engine.")

    run = subparsers.add_parser("run", help="Run a code.")
    run.add_argument("--name", dest='script_name',
                     type=str,
                     default=None,
                     help="Name of the script to run.")

    rundriver = subparsers.add_parser("rundriver", help="Run a calculation with a test driver.")
    rundriver.add_argument("--name", dest='driver_name',
                           type=str,
                           default=None,
                           help="Name of the test driver to run.")

    interactive = subparsers.add_parser("interactive", help="Start an interactive session within a container.")
    interactive.add_argument("--image", dest='image_name',
                             type=str,
                             default=None,
                             help="Name of the image to run interactively.")

    args = vars(parser.parse_args())
    if args["command"] is None:
        parser.print_help(sys.stderr)
        exit(1)

    return args

def main(args=None):
    if args is None:
        args = parse_args()

    command = args.pop("command")
    if command == "build":
        commands.command_build( args )
    elif command == "report":
        commands.command_report()
    elif command == "startproject":
        commands.command_startproject( args )
    elif command == "run":
        commands.command_run( args )
    elif command == "rundriver":
        commands.command_rundriver( args )
    elif command == "interactive":
        commands.command_interactive( args )



import os
import subprocess
from .utils import utils as ut

def install_all( base_path, build_type ):
    # Read the yaml file
    mdimechanic_yaml = ut.get_mdimechanic_yaml( base_path )

    build_release = False
    if build_type == "release":
       build_release = True

    # Read the script to build the image from the yaml file
    build_image_lines = mdimechanic_yaml['docker']['build_image']
    build_image_script = "#!/bin/bash -l\nset -e\n"
    for line in build_image_lines:
        build_image_script += line + '\n'

    if build_release:
        # Also build the engine within this image
        build_engine_lines = ""
        if 'build_code' in mdimechanic_yaml['docker']:
            build_engine_lines = mdimechanic_yaml['docker']['build_code']
        else:
            build_engine_lines = mdimechanic_yaml['docker']['build_engine']
        build_image_script += "mkdir -p /repo\n"
        build_image_script += "cd /repo\n"
        for line in build_engine_lines:
            build_image_script += line + '\n'
        #build_image_script += "cd /\n"
        #build_image_script += "mv /repo /release"

    # Write the script to build the image
    image_script_path = os.path.join( base_path, "docker", ".temp", "build_image.sh" )
    os.makedirs(os.path.dirname(image_script_path), exist_ok=True)
    ut.write_as_bytes( build_image_script, image_script_path )

    # Read the script to build the engine from the yaml file
    build_engine_lines = ""
    if 'build_code' in mdimechanic_yaml['docker']:
        build_engine_lines = mdimechanic_yaml['docker']['build_code']
    else:
        build_engine_lines = mdimechanic_yaml['docker']['build_engine']
    build_engine_script = "#!/bin/bash -l\nset -e\ncd /repo\n"
    for line in build_engine_lines:
        build_engine_script += line + '\n'

    # Write the script to build the engine
    engine_script_path = os.path.join( base_path, ".mdimechanic", ".temp", "build_engine.sh" )
    os.makedirs(os.path.dirname(engine_script_path), exist_ok=True)
    ut.write_as_bytes( build_engine_script, engine_script_path )

    # Switch to the package directory
    package_path = ut.get_package_path()
    os.chdir(package_path)

    # Build the MDI base image
    ret = os.system("docker build -t mdi/base mdimechanic/docker/base")
    if ret != 0:
        raise Exception("Unable to build the MDI Mechanic image")

    # Build the MDI Mechanic image
    ret = os.system("docker build -t mdi_mechanic/mdi_mechanic mdimechanic/docker")
    if ret != 0:
        raise Exception("Unable to build the MDI Mechanic image")

    # Switch to the base directory
    os.chdir(base_path)

    # Generate ssh keys within the MDI Mechanic image
    ssh_proc = subprocess.Popen( ["docker", "run", "--rm",
                                    "-v", str(base_path) + ":/repo",
                                    "-v", str(package_path) + ":/MDI_Mechanic",
                                    "mdi_mechanic/mdi_mechanic",
                                    "bash", "-c",
                                    "cd /MDI_Mechanic/mdimechanic/docker/ssh && bash ../../utils/generate_ssh_keys.sh"],
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    ssh_tup = ssh_proc.communicate()
    if ssh_proc.returncode != 0:
        ut.docker_error( ssh_tup, "Error during ssh key generation." )

    # Build the engine image
    tag = mdimechanic_yaml['docker']['image_name']
    if build_release:
        tag += ":release"
    else:
        tag += ":dev"
    build_command = "docker build -t " + tag + " docker"
    ret = os.system( build_command )
    if ret != 0:
        raise Exception("Unable to build the engine image")

    # This script will be executed on entry to the image
    build_entry_script = '''#!/bin/bash -l
set -e
cd /repo
bash .mdimechanic/.temp/build_engine.sh
'''

    # Write the entry script into the mounted volume
    build_entry_path = os.path.join( base_path, "docker", ".temp", "build_entry.sh" )
    os.makedirs(os.path.dirname(build_entry_path), exist_ok=True)
    ut.write_as_bytes( build_entry_script, build_entry_path )

    # Check if there are any custom GPU options in mdimechanic.yml
    gpu_options = ""
    if 'gpu' in mdimechanic_yaml['docker']:
        gpu_options = " --gpus all"

    if not build_release:

        # Build the engine, within its Docker image
        docker_string = "docker run --rm" + str(gpu_options) + " -v " + str(base_path) + ":/repo -v " + str(package_path) + ":/MDI_Mechanic " + mdimechanic_yaml['docker']['image_name'] + ":dev bash -l /repo/docker/.temp/build_entry.sh"
        ret = os.system(docker_string)
        if ret != 0:
            raise Exception("Unable to build the engine")



import os
import subprocess
import shutil
from .utils.utils import format_return, insert_list, docker_error, get_mdi_standard, get_compose_path, get_package_path, get_mdimechanic_yaml, write_as_bytes


def run( script_name, base_path ):
    mdimechanic_yaml = get_mdimechanic_yaml( base_path )

    # Get the path to the docker-compose file
    docker_path = os.path.join( base_path, ".mdimechanic", ".temp" )

    # Get a list of all containers in this calculation
    containers = [ key for key in mdimechanic_yaml['run_scripts'][script_name]['containers'] ]

    # Create the docker-compose.yml file
    docker_compose_text='''version: '3'

services:
'''

    for icontainer in range(len(containers)):
        docker_compose_text += f'  {containers[icontainer]}:'

        container_yaml = mdimechanic_yaml['run_scripts'][script_name]['containers'][containers[icontainer]]

        if not 'image' in container_yaml:
            raise Exception("No image was provided for container \"" + str(containers[icontainer]) + "\".  Please provide the image name in mdimechanic.yml.")

        image_name = container_yaml['image']
        script_file_name = "docker_mdi_" + str(icontainer) + ".sh"

        textargs = {'image_name': container_yaml['image'],
                    'workdir': base_path,
                    'packagedir': get_package_path(),
                    'script_file_name': script_file_name,
                    'icontainer': icontainer,
                    'container_name': containers[icontainer]}
        docker_compose_text += '''
    image: "{image_name}"
    command: bash -l -c "bash -l /mdi_shared/.mdimechanic/.temp/{script_file_name}"
    volumes:
      - '{workdir}:/mdi_shared'
      - '{packagedir}:/MDI_Mechanic'
    networks:
      mdinet:
        aliases:
          - {container_name}
'''.format(**textargs)

        if 'gpu' in mdimechanic_yaml['docker']:
                docker_compose_text += '''deploy:
      resources:
        reservations:
          devices:
            - driver: nvidia
              count: all
              capabilities: [gpu]
'''

    docker_compose_text += '''
networks:
  mdinet:
    driver: "bridge"
'''

    docker_compose_path = os.path.join( base_path, ".mdimechanic", ".temp", "docker-compose.yml" )
    os.makedirs(os.path.dirname(docker_compose_path), exist_ok=True)
    write_as_bytes( docker_compose_text, docker_compose_path )

    # Write the run script for each of the engines
    for icontainer in range(len(containers)):

        container_yaml = mdimechanic_yaml['run_scripts'][script_name]['containers'][containers[icontainer]]

        # Write the run script for the engine
        script_lines = container_yaml['script']
        script = '''#!/bin/bash -l\nset -e
cd /mdi_shared
export MDI_OPTIONS=\'-role ENGINE -name TESTCODE -method TCP -hostname mdi_mechanic -port 8021\'
if [ ! -d "/repo" ]; then
    ln -s /mdi_shared /repo
fi
'''
        #script = "#!/bin/bash\nset -e\ncd /mdi_shared\n"
        #script += "export MDI_OPTIONS=\'-role ENGINE -name TESTCODE -method TCP -hostname mdi_mechanic -port 8021\'\n"
        for line in script_lines:
            script += line + '\n'

        # Write the script to run the test
        script_file_name = "docker_mdi_" + str(icontainer) + ".sh"
        script_path = os.path.join( base_path, ".mdimechanic", ".temp", script_file_name )
        os.makedirs(os.path.dirname(script_path), exist_ok=True)
        write_as_bytes( script, script_path )

    # Launch with docker-compose
    docker_env = os.environ
    up_proc = subprocess.Popen( ["docker-compose", "up"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                cwd=docker_path, env=docker_env )
    up_tup = up_proc.communicate()
    up_out = format_return(up_tup[0])
    up_err = format_return(up_tup[1])

    # Run "docker-compose down"
    down_proc = subprocess.Popen( ["docker-compose", "down"],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                  cwd=docker_path, env=docker_env )
    down_tup = down_proc.communicate()
    down_out = format_return(down_tup[0])
    down_err = format_return(down_tup[1])

    if up_proc.returncode != 0:
        docker_error( up_tup, "Driver test returned non-zero exit code." )

    elif down_proc.returncode != 0:
        docker_error( down_tup, "Driver test returned non-zero exit code on docker down." )

    else:
        print("====================================================")
        print("================ Output from Docker ================")
        print("====================================================")
        print(up_out)
        print("====================================================")
        print("============== End Output from Docker ==============")
        print("====================================================")
